==> turbina-2.05.tgz: turbina-2.05/src/readmass.cpp <==
#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include "readmass.h"
#include <map>

// added 22.09.2003 to access standard C++ types without std::
using namespace std;

#ifdef READMASSTEST


int main(int argc, const char * argv[] ) {
	FILE * f1 = fopen("mass.in","rt");
	FILE * f2 = fopen("stm.in","rt");
	char type;
	while (type = rdm::readline(f1,'I',f2)) {
		cout << "T="<<type;
		int h,m,s;
		if (!rdm::isheader()) cout << " UT="<<rdm::getut()<<' ';
		switch(type) {
			case 'P':
				cout << "Par="<<rdm::getname()<<" Val="<<rdm::getvalue()<<'\n';
			break; case 'M':
				cout <<"Mode="<<rdm::getname();
				if (rdm::getvalue().length())
					cout <<" AirM="<<rdm::getvalue();
				cout <<'\n';
			break; case 'i': case 'j':
				cout << "Addr=" <<rdm::getaddr()<<'\n';
			break; case 'O':
				cout << "HR="<<rdm::getname()<<" is "<<rdm::getvalue()<<'\n';
			break; default:
				cout << rdm::getvalue() << '\n';
		}
	}

	int cnt = rdm::grabparams(f1);
	if (!cnt) {
		cerr << "No parameters found in mass-file\n";
	} else
		for (;;) {
			cout << cnt <<
				"'P'-records\nInput parameter name or [Enter] to exit:";
			string pname ;
			getline(cin,pname);
			if (!pname.length()) break;
			cout << pname << '=' << rdm::getvalue(pname) << '\n';
		}

	if (f1) fclose(f1);
	if (f2) fclose(f2);
	return 0;
}
#endif

namespace rdm {
	std::map<string,string> parmap;
	// line content storages: see documentation in readmass.h
	string value;
	string name;
	string name_org;
	long addr = 0;
	char type = 0;
	bool ishead = false;
	long lastfpos = 0;	// mass-file position BEFORE readling line
	int uyear = -1;
	int umonth = -1;
	int uday = -1;
	int uth = -1;
	int utm = -1;
	int uts = -1;

	const int LONGEST = 500;	// longest mass-line length
	char buffer[LONGEST];		// reading buffer
	const int UTLEN = 8;	// UT representation length
	const int DATEUTLEN = 10+1+UTLEN;	// DATE+UT representation length
	const int LENBEG = 1+1+UTLEN;	// shortest mass-line length (prefix+UT)
	bool swed = false;   // read stm if true
}

char rdm::readline(FILE * mf) {
	if (!mf) return 0;
	lastfpos = ftell(mf) ;
	if (!fgets(buffer, LONGEST, mf) || ((int)strlen(buffer)<LENBEG)) return 0;

	// remove backspace character:
	char * cr = strrchr(buffer,'\n');
	if (cr) {
		// overwrite with a end-character:
		*cr = 0;
	}
	// get prefix:
	type = buffer[0];
	ishead = false;

	if (type=='#') {
		// comment or header line?
		ishead = !isspace(buffer[1]);

		if (ishead) {
			value = buffer;
			return type;
		}
	}
	char * rest = &(buffer[1]);
	while (*rest && isspace(*rest)) rest++; // prefix and space skipped

	int h=99,m=99,s=99;
	// version 1.6: date components:
	int yy=0,mm=0,dd=0;
	// first try to extract full date, if fails - the time only:
	if (sscanf(rest,"%d-%d-%d %d:%d:%d",&yy,&mm,&dd,&h,&m,&s)!=6) {
		if (sscanf(rest,"%d:%d:%d",&h,&m,&s)!=3) return 0;
		rest += UTLEN; // UT record skipped
	} else {
		// validate date:
		if (yy<2001 || yy>2100 || mm<1 || mm>12 || dd<1 || dd>31) return 0;
		uyear = yy; umonth = mm; uday = dd;
		rest += DATEUTLEN; // DATE+UT record skipped
	}
	// validate time: we allow m=60 and s=60 (trbtime.cpp bug)
	if (h<0 || h>=100 || m<0 || m>60 || s<0 || s>60) return 0;
	uth = h; utm = m; uts = s;

	while (*rest && isspace(*rest)) rest++; // spaces skipped

	if (!strlen(rest)) return 0;	// stop if empty string to the right of UT

	char * c;
	switch (type) {
		case 'P' :
			c = strchr(rest,'=');
			if (!c) return 0;
			*c = 0;	// set the terminate character
			value = c+1;	// all after first '='
			// turn the name into the upper case:
			name_org = rest;
			for (char * ch=rest; (*ch)!=0; ch++) *ch = toupper(*ch);
			name = rest;	// all before '='
		break;
		case 'M' :
			// turn the name into the upper case:
			value = "";
			// locate the possibly existing number at the end:
			// skip the spaces at the tail and cancel them:
			c=rest+strlen(rest)-1;
			while ( isspace(*c) && c!=rest) {*c=0; c--;}
			// skip non-spaces before this:
			if (c!=rest) while (!isspace(*c) && c!=rest) c-- ;
			// skip the spaces before the last word:
			if (c!=rest) while ( isspace(*c) && c!=rest) c-- ;
			if (c!=rest) { 	
				// still not the beginning: indeed there are several words.
				// Go back to the first space before the last word:
				c++ ;
				// try to convert the last word into the number:
				char * endptr;
				strtod(c,&endptr);
				if (endptr!=c) {
					// successful conversion: indeed number: isolate it in the "value":
					*c = 0;	// add a terminator for "name" on the 1st space after it
					// start value at the last word:
					do c++; while ( isspace(*c) && (c!=0));
					value = c;
				} 
			}
			name_org = rest;
			for (char * ch=rest; (*ch)!=0; ch++) *ch = toupper(*ch);
			name = rest;
		break;
		case 'O':
			// find and return the first single word (normally HR number):
			c = rest;
			while (*c && !isspace(*c)) c++; // non-spaces skipped
			if (c!=rest) {
				*c = 0;
				name = name_org = rest;
				c++;
				while (*c && isspace(*c)) c++; // spaces skipped
				value = c;
//printf("\tHR=%s, Pars=%s\n",name.c_str(),value.c_str());
			}
		break;
		case 'i': case 'j':
			// find and return the address, or -1 if not written
			c = strchr(rest,'@');
			if (!c)
				addr = NOADDR;
			else {
				*c = 0;	// terminate string for value here
				char * endptr;
				addr = strtol(c+1,&endptr,16);
				if (endptr==c+1) addr = NOADDR;	// error of recognition
//printf("\tAddr=%ld\n",addr);
			}
		// no break here: we need the rest of string as "value"
		default:
			value = rest;
//printf("\tRest=%s\n",value.c_str());
	}
	return type;
}

char rdm::gettype() { return type; }

long rdm::getstartpos() { return lastfpos; }

bool rdm::isheader() { return ishead; }

double rdm::getut(int * uuth, int * uutm, int * uuts) {
	if (uuth)
		*uuth = uth;
	if (uutm)
		*uutm = utm;
	if (uuts)
		*uuts = uts;
	if (uth>=0)
		return uth + utm/60. + uts / 3600.;
	else
		return -1;
}

double rdm::getut(int * uuyear, int * uumonth, int * uuday, int * uuth, int * uutm, int * uuts) {
	if (uuyear) *uuyear = uyear;
	if (uumonth) *uumonth = umonth;
	if (uuday) *uuday = uday;
	return rdm::getut(uuth, uutm, uuts);
}

string rdm::getut() {
	const char UTFMT[] = "%02d:%02d:%02d";
	const char DATEUTFMT[] = "%04d-%02d-%02d %02d:%02d:%02d";
	const int LDATEUTFMT = 10+1+8;
	static char ut[LDATEUTFMT] = " ";
	if (uyear>0) {
		sprintf(ut,DATEUTFMT,uyear,umonth,uday,uth,utm,uts);
	} else {
		sprintf(ut,UTFMT,uth,utm,uts);
	}
	return string(ut);
}

string rdm::getname(bool org) { if (org) return name_org; else return name; }

string rdm::getvalue(const string pname) {
	if (pname.length()) {
		return parmap[pname];
		// access the parameters map:
	} else {
		return value;
	}
}

long rdm::getaddr() { return addr; }

int rdm::add2map(const string pname, const string pvalue) {
	if (pname.length())
		parmap[pname] = pvalue;
	return parmap.size() ;
}

int rdm::grabparams(FILE * mf) {
	int cnt = 0;
	char t = 0;
	parmap.clear();
	if (mf) fseek(mf,0,SEEK_SET);
	while ((t = readline(mf))!=0) {
		if (t=='P' && name.length() && value.length()) {
			// add to the parameter map:
			cnt++;
			add2map(name,value) ;
		}
	}
	return cnt;
}


char rdm::readline(FILE * mass, const char* sw, FILE * stm) {
	static string UTbeg, UTend;
 
	// read current line:
	char res = rdm::readline(rdm::swed && (stm!=0) ? stm : mass);
 
	if( rdm::swed && stm!=0 ) {
		string UT;
		// skip predecessing lines (assumes sorted stm-file!)
		while( ((UT = rdm::getut())<UTbeg) && (res = rdm::readline(stm)) );
		// now: either EOF or UT>=UTbeg:
		if (!res || (UT>UTend)) {
			// no lines left or found: scope over or empty UTbeg-UTend selection
			rdm::swed = false;
			// get back the last-read line if read:
			if (res) fseek(stm,rdm::getstartpos(),SEEK_SET);
			// read back the primary file:
			res = rdm::readline(mass) ;
		} // else - Ok: the line's read.
	} else {
		if( strchr(sw,res) ) {
			// switcher is encountered: check last field
			// check that there is UTend in the end:
			UTend = rdm::getvalue();
			// position of the beginning of UTend:
			int utEndPos = UTend.rfind(':') - 5 - 11;
			// locate "UTend":
			if (utEndPos > 0) {
				UTend = UTend.substr(utEndPos,10+1+8);
				// UTbeg assignment:
				UTbeg = rdm::getut();
//cout << UTbeg << "?" << UTend << "===" << utEndPos << '\n';
				// validate UTend relative to UTbeg:
				if ((UTend[2] == UTend[5]) &&
					// the use of character comparison is valid here:
				(UTend>=UTbeg)) {
					// Ok: limits are set.
					rdm::swed = true;
//cout << "UTbeg=" << UTbeg << " UTend=" << UTend << " swed=" << rdm::swed << '\n';
				}
			} //else:  old or invalid mass-file data: no switching
		}
	}
	return res;
}

==> turbina-2.056.tgz: turbina-2.056/src/readmass.cpp <==
#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include "readmass.h"
#include <map>
#include <iostream>
// added 22.09.2003 to access standard C++ types without std::
using namespace std;

#ifdef READMASSTEST


int main(int argc, const char * argv[] ) {
	FILE * f1 = fopen("mass.in","rt");
	FILE * f2 = fopen("stm.in","rt");
	char type;
	while (type = rdm::readline(f1,'I',f2)) {
		cout << "T="<<type;
		int h,m,s;
		if (!rdm::isheader()) cout << " UT="<<rdm::getut()<<' ';
		switch(type) {
			case 'P':
				cout << "Par="<<rdm::getname()<<" Val="<<rdm::getvalue()<<'\n';
			break; case 'M':
				cout <<"Mode="<<rdm::getname();
				if (rdm::getvalue().length())
					cout <<" AirM="<<rdm::getvalue();
				cout <<'\n';
			break; case 'i': case 'j':
				cout << "Addr=" <<rdm::getaddr()<<'\n';
			break; case 'O':
				cout << "HR="<<rdm::getname()<<" is "<<rdm::getvalue()<<'\n';
			break; default:
				cout << rdm::getvalue() << '\n';
		}
	}

	int cnt = rdm::grabparams(f1);
	if (!cnt) {
		cerr << "No parameters found in mass-file\n";
	} else
		for (;;) {
			cout << cnt <<
				"'P'-records\nInput parameter name or [Enter] to exit:";
			string pname ;
			getline(cin,pname);
			if (!pname.length()) break;
			cout << pname << '=' << rdm::getvalue(pname) << '\n';
		}

	if (f1) fclose(f1);
	if (f2) fclose(f2);
	return 0;
}
#endif

namespace rdm {
	std::map<string,string> parmap;
	// line content storages: see documentation in readmass.h
	string value;
	string name;
	string name_org;
	long addr = 0;
	char type = 0;
	bool ishead = false;
	long lastfpos = 0;	// mass-file position BEFORE readling line
	int uyear = -1;
	int umonth = -1;
	int uday = -1;
	int uth = -1;
	int utm = -1;
	int uts = -1;

	const int LONGEST = 500;	// longest mass-line length
	char buffer[LONGEST];		// reading buffer
	const int UTLEN = 8;	// UT representation length
	const int DATEUTLEN = 10+1+UTLEN;	// DATE+UT representation length
	const int LENBEG = 1+1+UTLEN;	// shortest mass-line length (prefix+UT)
	bool swed = false;   // read stm if true
}

char rdm::readline(FILE * mf) {
	if (!mf) return 0;
	lastfpos = ftell(mf) ;
	if (!fgets(buffer, LONGEST, mf) || ((int)strlen(buffer)<LENBEG)) return 0;

	// remove backspace character:
	char * cr = strrchr(buffer,'\n');
	if (cr) {
		// overwrite with a end-character:
		*cr = 0;
	}
	// get prefix:
	type = buffer[0];
	ishead = false;

	if (type=='#') {
		// comment or header line?
		ishead = !isspace(buffer[1]);

		if (ishead) {
			value = buffer;
			return type;
		}
	}
	char * rest = &(buffer[1]);
	while (*rest && isspace(*rest)) rest++; // prefix and space skipped

	int h=99,m=99,s=99;
	// version 1.6: date components:
	int yy=0,mm=0,dd=0;
	// first try to extract full date, if fails - the time only:
	if (sscanf(rest,"%d-%d-%d %d:%d:%d",&yy,&mm,&dd,&h,&m,&s)!=6) {
		if (sscanf(rest,"%d:%d:%d",&h,&m,&s)!=3) return 0;
		rest += UTLEN; // UT record skipped
	} else {
		// validate date:
		if (yy<2001 || yy>2100 || mm<1 || mm>12 || dd<1 || dd>31) return 0;
		uyear = yy; umonth = mm; uday = dd;
		rest += DATEUTLEN; // DATE+UT record skipped
	}
	// validate time: we allow m=60 and s=60 (trbtime.cpp bug)
	if (h<0 || h>=100 || m<0 || m>60 || s<0 || s>60) return 0;
	uth = h; utm = m; uts = s;

	while (*rest && isspace(*rest)) rest++; // spaces skipped

	if (!strlen(rest)) return 0;	// stop if empty string to the right of UT

	char * c;
	switch (type) {
		case 'P' :
			c = strchr(rest,'=');
			if (!c) return 0;
			*c = 0;	// set the terminate character
			value = c+1;	// all after first '='
			// turn the name into the upper case:
			name_org = rest;
			for (char * ch=rest; (*ch)!=0; ch++) *ch = toupper(*ch);
			name = rest;	// all before '='
		break;
		case 'M' :
			// turn the name into the upper case:
			value = "";
			// locate the possibly existing number at the end:
			// skip the spaces at the tail and cancel them:
			c=rest+strlen(rest)-1;
			while ( isspace(*c) && c!=rest) {*c=0; c--;}
			// skip non-spaces before this:
			if (c!=rest) while (!isspace(*c) && c!=rest) c-- ;
			// skip the spaces before the last word:
			if (c!=rest) while ( isspace(*c) && c!=rest) c-- ;
			if (c!=rest) { 	
				// still not the beginning: indeed there are several words.
				// Go back to the first space before the last word:
				c++ ;
				// try to convert the last word into the number:
				char * endptr;
				strtod(c,&endptr);
				if (endptr!=c) {
					// successful conversion: indeed number: isolate it in the "value":
					*c = 0;	// add a terminator for "name" on the 1st space after it
					// start value at the last word:
					do c++; while ( isspace(*c) && (c!=0));
					value = c;
				} 
			}
			name_org = rest;
			for (char * ch=rest; (*ch)!=0; ch++) *ch = toupper(*ch);
			name = rest;
		break;
		case 'O':
			// find and return the first single word (normally HR number):
			c = rest;
			while (*c && !isspace(*c)) c++; // non-spaces skipped
			if (c!=rest) {
				*c = 0;
				name = name_org = rest;
				c++;
				while (*c && isspace(*c)) c++; // spaces skipped
				value = c;
//printf("\tHR=%s, Pars=%s\n",name.c_str(),value.c_str());
			}
		break;
		case 'i': case 'j':
			// find and return the address, or -1 if not written
			c = strchr(rest,'@');
			if (!c)
				addr = NOADDR;
			else {
				*c = 0;	// terminate string for value here
				char * endptr;
				addr = strtol(c+1,&endptr,16);
				if (endptr==c+1) addr = NOADDR;	// error of recognition
//printf("\tAddr=%ld\n",addr);
			}
		// no break here: we need the rest of string as "value"
		default:
			value = rest;
//printf("\tRest=%s\n",value.c_str());
	}
	return type;
}

char rdm::gettype() { return type; }

long rdm::getstartpos() { return lastfpos; }

bool rdm::isheader() { return ishead; }

double rdm::getut(int * uuth, int * uutm, int * uuts) {
	if (uuth)
		*uuth = uth;
	if (uutm)
		*uutm = utm;
	if (uuts)
		*uuts = uts;
	if (uth>=0)
		return uth + utm/60. + uts / 3600.;
	else
		return -1;
}

double rdm::getut(int * uuyear, int * uumonth, int * uuday, int * uuth, int * uutm, int * uuts) {
	if (uuyear) *uuyear = uyear;
	if (uumonth) *uumonth = umonth;
	if (uuday) *uuday = uday;
	return rdm::getut(uuth, uutm, uuts);
}

string rdm::getut() {
	const char UTFMT[] = "%02d:%02d:%02d";
	const char DATEUTFMT[] = "%04d-%02d-%02d %02d:%02d:%02d";
	const int LDATEUTFMT = 10+1+9;
	static char ut[LDATEUTFMT] = " ";
	if (uyear>0) {
		sprintf(ut,DATEUTFMT,uyear,umonth,uday,uth,utm,uts);
	} else {
		sprintf(ut,UTFMT,uth,utm,uts);
	}
	return string(ut);
}

string rdm::getname(bool org) { if (org) return name_org; else return name; }

string rdm::getvalue(const string pname) {
	if (pname.length()) {
		return parmap[pname];
		// access the parameters map:
	} else {
		return value;
	}
}

long rdm::getaddr() { return addr; }

int rdm::add2map(const string pname, const string pvalue) {
	if (pname.length())
		parmap[pname] = pvalue;
	return parmap.size() ;
}

int rdm::grabparams(FILE * mf) {
	int cnt = 0;
	char t = 0;
	parmap.clear();
	if (mf) fseek(mf,0,SEEK_SET);
	while ((t = readline(mf))!=0) {
		if (t=='P' && name.length() && value.length()) {
			// add to the parameter map:
			cnt++;
			add2map(name,value) ;
		}
	}
	return cnt;
}


char rdm::readline(FILE * mass, const char* sw, FILE * stm) {
	static string UTbeg, UTend;
 
	// read current line:
	char res = rdm::readline(rdm::swed && (stm!=0) ? stm : mass);
 
	if( rdm::swed && stm!=0 ) {
		string UT;
		// skip predecessing lines (assumes sorted stm-file!)
		while( ((UT = rdm::getut())<UTbeg) && (res = rdm::readline(stm)) );
		// now: either EOF or UT>=UTbeg:
		if (!res || (UT>UTend)) {
			// no lines left or found: scope over or empty UTbeg-UTend selection
			rdm::swed = false;
			// get back the last-read line if read:
			if (res) fseek(stm,rdm::getstartpos(),SEEK_SET);
			// read back the primary file:
			res = rdm::readline(mass) ;
		} // else - Ok: the line's read.
	} else {
		if( strchr(sw,res) ) {
			// switcher is encountered: check last field
			// check that there is UTend in the end:
			UTend = rdm::getvalue();
			// position of the beginning of UTend:
			size_t lutbeg = getut().length();
			int utEndPos = UTend.rfind(':') + 3 - lutbeg;
			// locate "UTend":
			if (utEndPos > 0) {
				UTend = UTend.substr(utEndPos,lutbeg);
				// UTbeg assignment:
				UTbeg = rdm::getut();
//cout << UTbeg << "?" << UTend << "===" << utEndPos << '\n';
				// validate UTend relative to UTbeg:
				if ((UTend[lutbeg-3] == UTend[lutbeg-6]) &&
					// the use of character comparison is valid here:
				(UTend>=UTbeg)) {
					// Ok: limits are set.
					rdm::swed = true;
//cout << "UTbeg=" << UTbeg << " UTend=" << UTend << " swed=" << rdm::swed << '\n';
				}
			} //else:  old or invalid mass-file data: no switching
		}
	}
	return res;
}
